#include <WiFi.h>
#include <PubSubClient.h>
#include <FastLED.h>
#include <AccelStepper.h>

int garage_r = 255;
int garage_g = 255;
int garage_b = 255;

int roof_r = 255;
int roof_g = 255;
int roof_b = 255;

int bedroom_r = 255;
int bedroom_g = 255;
int bedroom_b = 255;

int bathroom_r = 255;
int bathroom_g = 255;
int bathroom_b = 255;



// Define step constants
const int stepsPerRevolution = 2038; // change this to fit the number of steps per revolution

#define STEPPER_IN1 18
#define STEPPER_IN2 5
#define STEPPER_IN3 17
#define STEPPER_IN4 16


#define STEPPER2_IN1 26
#define STEPPER2_IN2 27
#define STEPPER2_IN3 14
#define STEPPER2_IN4 12

#define home_switch 39


long TravelX;  // Used to store the X value entered in the Serial Monitor
int move_finished = 1; // Used to check if move is completed
long initial_homing = -1; // Used to Home Stepper at startup

AccelStepper garageStepper(4, STEPPER_IN1, STEPPER_IN2, STEPPER_IN3, STEPPER_IN4);
AccelStepper elevatorStepper(4, STEPPER2_IN1, STEPPER2_IN2, STEPPER2_IN3, STEPPER2_IN4);


int stepCount = 0; // number of steps the motor has taken


const char* ssid     = "RPiHotspotN";
const char* password = "1234567890";
const char* mqttServer = "192.168.50.5";

WiFiClient espClient;
PubSubClient client(espClient);

// How many leds in your strip?
#define NUM_LEDS 5
#define DATA_PIN 22
// Define the array of leds
CRGB leds[NUM_LEDS];

bool garageDoorOpen = false;
bool garageDoorClose = false;

bool goToGarage = false;
bool goToBedroom = false;
bool goToBathroom = false;
bool goToRooftop = false;

volatile unsigned long last_micros;


int garageLedButton = 4; //brown
int garageLedState = LOW;
int garageLedButtonState;
int garageLedLastButtonState = LOW;

int bedroomLedButton = 15;// purple
int bedroomLedState = LOW;
int bedroomLedButtonState;
int bedroomLedLastButtonState = LOW;


int bathroomLedButton = 2; //red
int bathroomLedState = LOW;
int bathroomLedButtonState;
int bathroomLedLastButtonState = LOW;

int rooftopLedButton = 13; //black
int rooftopLedState = LOW;
int rooftopLedButtonState;
int rooftopLedLastButtonState = LOW;

unsigned long lastDebounceTime = 0;
unsigned long debounceDelay = 50;

void setup() {
  Serial.begin(115200);

  FastLED.addLeds<WS2812, DATA_PIN, GRB>(leds, NUM_LEDS);
  pinMode(garageLedButton, INPUT);
  pinMode(bedroomLedButton, INPUT);
  pinMode(bathroomLedButton, INPUT);
  pinMode(rooftopLedButton, INPUT);

  garageStepper.setCurrentPosition(0);  // Set the current position as zero for now
  garageStepper.setMaxSpeed(1000.0);      // Set Max Speed of Stepper (Slower to get better accuracy)
  garageStepper.setAcceleration(200);  // Set Acceleration of Stepper

    setup_elevator();

  setup_wifi_and_mqtt();

}

void setup_elevator() {
  pinMode(home_switch, INPUT);

  elevatorStepper.setMaxSpeed(1000.0);      // Set Max Speed of Stepper (Slower to get better accuracy)
  elevatorStepper.setAcceleration(200);  // Set Acceleration of Stepper


  Serial.print("Stepper is Homing . . . . . . . . . . . ");

  elevatorStepper.setCurrentPosition(0);  // Set the current position as zero for now
  elevatorStepper.setMaxSpeed(1000.0);      // Set Max Speed of Stepper (Slower to get better accuracy)
  elevatorStepper.setAcceleration(200);  // Set Acceleration of Stepper
  initial_homing = 1;
  while (!digitalRead(home_switch))  {

    elevatorStepper.moveTo(initial_homing);
    elevatorStepper.run();
    initial_homing--;
    delay(5);
  }
  elevatorStepper.setCurrentPosition(0);
  Serial.println("Homing Completed");
  Serial.println(elevatorStepper.currentPosition());
  elevatorStepper.setMaxSpeed(1000);      // Set Max Speed of Stepper (Faster for regular movements)
  elevatorStepper.setAcceleration(200);  // Set Acceleration of Stepper
}

void setup_wifi_and_mqtt() {
  delay(10);
  unsigned long startedWaiting = millis();
  // We start by connecting to a WiFi network
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);

  WiFi.begin(ssid, password);

  int trial = 0;
  while (WiFi.status() != WL_CONNECTED && millis() - startedWaiting <= 10000) {
    delay(500);
    Serial.print(".");
    trial++;
    if (trial == 5) {
      break;
    }
  }

  if (WiFi.status() == WL_CONNECTED) {
    Serial.println("");
    Serial.println("WiFi connected");
    Serial.println("IP address: ");
    Serial.println(WiFi.localIP());

    Serial.println("Connecting to MQTT...");

    trial = 0;
    client.setServer(mqttServer, 1883);
    client.setCallback(callback);

    while (!client.connected()) {
      Serial.print("Attempting MQTT connection...");
      trial++;
      // Attempt to connect
      if (client.connect("left-house")) {
        Serial.println("connected");

        // Subscribe
        client.subscribe("house/garage");
        client.subscribe("house/bedroom");
        client.subscribe("house/bathroom");
        client.subscribe("house/rooftop");
        client.subscribe("house/garage_color_change");
        client.subscribe("house/bedroom_color_change");
        client.subscribe("house/bathroom_color_change");
        client.subscribe("house/rooftop_color_change");

        client.subscribe("house/elevator_garage");
        client.subscribe("house/elevator_bedroom");
        client.subscribe("house/elevator_bathroom");
        client.subscribe("house/elevator_rooftop");


      } else {
        Serial.print("failed, rc=");
        Serial.print(client.state());
        Serial.println(" try again in 5 seconds");
        // Wait 5 seconds before retrying
        delay(5000);
      }

      if (trial == 5) {
        break;
      }
    }
  }

}


void callback(char* topic, byte* message, unsigned int length) {
  Serial.print("Message arrived on topic: ");
  Serial.print(topic);
  Serial.print(". Message: ");
  String messageTemp;

  for (int i = 0; i < length; i++) {
    Serial.print((char)message[i]);
    messageTemp += (char)message[i];
  }
  Serial.println();

  if (String(topic) == "house/garage") {
    if (messageTemp == "true") {
      leds[0] = CRGB(garage_r, garage_g, garage_b);
    } else if (messageTemp == "false") {
      leds[0] = CRGB(0, 0, 0);
    }
  }

  if (String(topic) == "house/bedroom") {
    if (messageTemp == "true") {
      leds[1] = CRGB(bedroom_r, bedroom_g, bedroom_b);
    } else if (messageTemp == "false") {
      leds[1] = CRGB(0, 0, 0);
    }
  }


  if (String(topic) == "house/bathroom") {
    if (messageTemp == "true") {
      leds[2] = CRGB(bathroom_r, bathroom_g, bathroom_b);
    } else if (messageTemp == "false") {
      leds[2] = CRGB(0, 0, 0);
    }
  }


  if (String(topic) == "house/rooftop") {
    if (messageTemp == "true") {
      leds[3] = CRGB(roof_r, roof_g, roof_b);
    } else if (messageTemp == "false") {
      leds[3] = CRGB(0, 0, 0);
    }
  }

  
  if (String(topic) == "house/garage_color_change") {
    long number = (long) strtol( &messageTemp[0], NULL, 16);
    garage_r = number >> 16;
    garage_g = number >> 8 & 0xFF;
    garage_b = number & 0xFF;
    leds[0] = CRGB(garage_r, garage_g, garage_b);
    client.publish("house/garage_switch_on", "1");
  }

  
  if (String(topic) == "house/bedroom_color_change") {
    long number = (long) strtol( &messageTemp[0], NULL, 16);
    bedroom_r = number >> 16;
    bedroom_g = number >> 8 & 0xFF;
    bedroom_b = number & 0xFF;
    leds[1] = CRGB(bedroom_r, bedroom_g, bedroom_b);
    client.publish("house/bedroom_switch_on", "1");
  }

  
  if (String(topic) == "house/bathroom_color_change") {
    long number = (long) strtol( &messageTemp[0], NULL, 16);
    bathroom_r = number >> 16;
    bathroom_g = number >> 8 & 0xFF;
    bathroom_b = number & 0xFF;
    leds[2] = CRGB(bathroom_r, bathroom_g, bathroom_b);
    client.publish("house/bathroom_switch_on", "1");
  }

  
  if (String(topic) == "house/rooftop_color_change") {
    long number = (long) strtol( &messageTemp[0], NULL, 16);
    roof_r = number >> 16;
    roof_g = number >> 8 & 0xFF;
    roof_b = number & 0xFF;
    leds[3] = CRGB(roof_r, roof_g, roof_b);
    client.publish("house/rooftop_switch_on", "1");
  }


  FastLED.show();

  goToGarage = false;
  goToBedroom = false;
  goToBathroom = false;
  goToRooftop = false;

  if (String(topic) == "house/elevator_garage") {
    goToGarage = true;
  } else if (String(topic) == "house/elevator_bedroom") {
    goToBedroom = true;
  } else if (String(topic) == "house/elevator_bathroom") {
    goToBathroom = true;
  } else if (String(topic) == "house/elevator_rooftop") {
    goToRooftop = true;
  }
}
//
//void reconnect() {
//  // Loop until we're reconnected
//  while (!client.connected()) {
//    Serial.print("Attempting MQTT connection...");
//    // Attempt to connect
//    if (client.connect("ESP8266Client")) {
//      Serial.println("connected");
//
//      // Subscribe
//      client.subscribe("house/garage");
//      client.subscribe("house/bedroom");
//      client.subscribe("house/bathroom");
//      client.subscribe("house/rooftop");
//
//      client.subscribe("house/elevator_garage");
//      client.subscribe("house/elevator_bedroom");
//      client.subscribe("house/elevator_bathroom");
//      client.subscribe("house/elevator_rooftop");
//
//
//
//    } else {
//      Serial.print("failed, rc=");
//      Serial.print(client.state());
//      Serial.println(" try again in 5 seconds");
//      // Wait 5 seconds before retrying
//      delay(5000);
//    }
//  }
//}
void loop() {
  //  if (!client.connected()) {
  //    reconnect();
  //  }

  int garageLedReading = digitalRead(garageLedButton);
  if (garageLedReading != garageLedLastButtonState) {
    lastDebounceTime = millis();
  }

  if ((millis() - lastDebounceTime) + debounceDelay) {
    if (garageLedReading != garageLedButtonState) {
      garageLedButtonState = garageLedReading;

      // only toggle the LED if the new button state is HIGH
      if (garageLedButtonState == HIGH) {
        garageLedState = !garageLedState;
        garageDoorOpen = false;
        garageDoorClose = false;
        if (garageLedState) {
          garageDoorOpen = true;
          leds[0] = CRGB(garage_r, garage_g, garage_b);
          client.publish("house/garage_switch_on", "1");
        } else {
          leds[0] = CRGB(0, 0, 0);
          garageDoorClose = true;
          client.publish("house/garage_switch_on", "0");
        }
      }
    }
  }

  garageLedLastButtonState = garageLedReading;


  int bedroomLedReading = digitalRead(bedroomLedButton);
  if (bedroomLedReading != bedroomLedLastButtonState) {
    lastDebounceTime = millis();
  }

  if ((millis() - lastDebounceTime) + debounceDelay) {
    if (bedroomLedReading != bedroomLedButtonState) {
      bedroomLedButtonState = bedroomLedReading;

      // only toggle the LED if the new button state is HIGH
      if (bedroomLedButtonState == HIGH) {
        Serial.println(bedroomLedReading);
        bedroomLedState = !bedroomLedState;
        if (bedroomLedState) {
          leds[1] = CRGB(bedroom_r, bedroom_g, bedroom_b);
          client.publish("house/bedroom_switch_on", "1");
        } else {
          leds[1] = CRGB(0, 0, 0);
          client.publish("house/bedroom_switch_on", "0");
        }
      }
    }
  }

  bedroomLedLastButtonState = bedroomLedReading;



  int bathroomLedReading = digitalRead(bathroomLedButton);
  if (bathroomLedReading != bathroomLedLastButtonState) {
    lastDebounceTime = millis();
  }

  if ((millis() - lastDebounceTime) + debounceDelay) {
    if (bathroomLedReading != bathroomLedButtonState) {
      bathroomLedButtonState = bathroomLedReading;

      // only toggle the LED if the new button state is HIGH
      if (bathroomLedButtonState == HIGH) {
        bathroomLedState = !bathroomLedState;
        if (bathroomLedState) {
          leds[2] = CRGB(bathroom_r, bathroom_g, bathroom_b);
          client.publish("house/bathroom_switch_on", "1");
        } else {
          leds[2] = CRGB(0, 0, 0);
          client.publish("house/bathroom_switch_on", "0");
        }
      }
    }
  }

  bathroomLedLastButtonState = bathroomLedReading;
  FastLED.show();


  int rooftopLedReading = digitalRead(rooftopLedButton);
  if (rooftopLedReading != rooftopLedLastButtonState) {
    lastDebounceTime = millis();
  }

  if ((millis() - lastDebounceTime) + debounceDelay) {
    if (rooftopLedReading != rooftopLedButtonState) {
      rooftopLedButtonState = rooftopLedReading;

      // only toggle the LED if the new button state is HIGH
      if (rooftopLedButtonState == HIGH) {
        rooftopLedState = !rooftopLedState;
        if (rooftopLedState) {
          leds[3] = CRGB(roof_r, roof_g, roof_b);
          client.publish("house/rooftop_switch_on", "1");
        } else {
          leds[3] = CRGB(0, 0, 0);
          client.publish("house/rooftop_switch_on", "0");
        }
      }
    }
  }

  rooftopLedLastButtonState = rooftopLedReading;
  FastLED.show();



  if (garageDoorOpen) {
    Serial.println(garageStepper.currentPosition());
    garageStepper.moveTo(1200);
    if ((garageStepper.distanceToGo() != 0)) {
      garageStepper.run();  // Move Stepper into position
    } else {
      garageDoorOpen = false;
    }
  } else if (garageDoorClose) {
    Serial.println(garageStepper.currentPosition());
    garageStepper.moveTo(0);
    if ((garageStepper.distanceToGo() != 0)) {
      garageStepper.run();  // Move Stepper into position
    } else {
      garageDoorClose = false;
    }
  }



  if (goToGarage) {
    Serial.println(elevatorStepper.currentPosition());
    elevatorStepper.moveTo(0);
    if ((elevatorStepper.distanceToGo() != 0)) {
      elevatorStepper.run();  // Move Stepper into position
    } else {
      goToGarage = false;
    }
  } else if (goToBedroom) {
    Serial.println(elevatorStepper.currentPosition());
    elevatorStepper.moveTo(1400);
    if ((elevatorStepper.distanceToGo() != 0)) {
      elevatorStepper.run();  // Move Stepper into position
    } else {
      goToBedroom = false;
    }
  } else if (goToBathroom) {
    Serial.println(elevatorStepper.currentPosition());
    elevatorStepper.moveTo(2800);
    if ((elevatorStepper.distanceToGo() != 0)) {
      elevatorStepper.run();  // Move Stepper into position
    } else {
      goToBathroom = false;
    }
  } else if (goToRooftop) {
    Serial.println(elevatorStepper.currentPosition());
    elevatorStepper.moveTo(4200 );
    if ((elevatorStepper.distanceToGo() != 0)) {
      elevatorStepper.run();  // Move Stepper into position
    } else {
      goToRooftop = false;
    }
  }




  client.loop();
}
